import shutil
from subprocess import check_call


class GCPBackend:
    def __init__(self):
        self._backend = 'gcp'
        self._call_prefix = ['gsutil', '-q', '-o', 'GSUtil:parallel_composite_upload_threshold=150M']

    def copy(self, recursive, parallel, ionice, filenames, quiet):
        call_args = ['ionice', '-c', '2', '-n', '7'] if ionice and (shutil.which('ionice') != None) else []
        call_args += self._call_prefix
        if parallel:
            call_args.append('-m')
        call_args.append('cp')
        if recursive:
            call_args.append('-r')
        call_args.extend(filenames)
        if not quiet:
            print(' '.join(call_args))
        check_call(call_args)

    def sync(self, parallel, ionice, source, target, quiet):
        # If target folder is local.
        if len(target.split('://')) == 1:
            import os
            if not os.path.exists(target):
                os.mkdir(target)

        call_args = ['ionice', '-c', '2', '-n', '7'] if ionice and (shutil.which('ionice') != None) else []
        call_args += self._call_prefix
        if parallel:
            call_args.append('-m')
        call_args.extend(['rsync', '-d', '-r', source, target])
        if not quiet:
            print(' '.join(call_args))
        check_call(call_args)

    def delete(self, recursive, parallel, filenames, quiet):
        call_args = self._call_prefix.copy()
        if parallel:
            call_args.append('-m')
        call_args.append('rm')
        if recursive:
            call_args.append('-r')
        call_args.extend(filenames)
        if not quiet:
            print(' '.join(call_args))
        check_call(call_args)

    def stat(self, filename):
        assert filename.startswith("gs://"), "Must be a GS URI!"
        is_folder = True if filename[-1]=='/' else False

        if is_folder:
            call_args = ['gsutil', '-q', 'stat', filename + '*']
        else:
            call_args = ['gsutil', '-q', 'stat', filename]

        check_call(call_args)



import argparse

example_text = """Examples:
  # AWS upload
  strato cp --backend aws file1 folder2 s3://my-bucket/target_folder/
  # AWS download
  strato cp --backend aws s3://my-bucket/source_folder/*.zip /target_folder/

  # GCP upload
  strato cp --backend gcp -m -r --ionice file1 folder2 gs://my-bucket/target_folder/
  # GCP download
  mkdir /target_folder
  strato cp --backend gcp -m gs://my-bucket/source_folder/*.zip /target_folder/

  # On local machine
  strato cp --backend local -r file1 folder2 /target_folder/
"""

def copy_files(backend, recursive, parallel, ionice, filenames, profile, quiet):
    assert backend in ['aws', 'gcp', 'local'], "Backend not supported!"

    if backend == 'aws':
        from strato.backends import AWSBackend
        be = AWSBackend()
        be.copy(ionice, filenames, profile, quiet)
    elif backend == 'gcp':
        from strato.backends import GCPBackend
        be = GCPBackend()
        be.copy(recursive, parallel, ionice, filenames, quiet)
    else:
        from strato.backends import LocalBackend
        be = LocalBackend()
        be.copy(recursive, ionice, filenames, quiet)


def main(argsv):
    parser = argparse.ArgumentParser(
        description="Copy files or folders. \nNotice that for AWS backend, a folder link must end with a slash '/'. \nFor GCP backend, if the target is a non-existing local path, you need to manually create it beforehand.",
        epilog=example_text,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument('--backend', dest='backend', action='store', required=True, help='Specify which backend to use. Available options: aws, gcp, local.')
    parser.add_argument('-r', dest='recursive', action='store_true', help="Recursive copy. Not needed for AWS backend.")
    parser.add_argument('-m', dest='parallel', action='store_true', help="Run operations in parallel. Only available for GCP backend.")
    parser.add_argument('--ionice', dest='ionice', action='store_true', help="Run with ionice to avoid monopolizing local disk's I/O. Only available for Linux.")
    parser.add_argument('--profile', dest='profile', type=str, action='store', help='AWS profile. Only works for aws backend, and use the default profile if not provided.')
    parser.add_argument('--quiet', dest='quiet', action='store_true', help="Hide the underlying cloud command.")
    parser.add_argument('files', metavar='filenames', type=str, nargs='+', help='List of file paths.')

    args = parser.parse_args(argsv)
    copy_files(args.backend, args.recursive, args.parallel, args.ionice, args.files, args.profile, args.quiet)



import os, shutil
import boto3
from subprocess import check_call, CalledProcessError


def parse_wildcard(filepath):
    prefix = "s3://" if filepath.startswith("s3://") else ""
    fp_list = filepath[5:].split('/') if prefix == "s3://" else filepath.split('/')
    wd_idx = -1
    for i in range(len(fp_list)):
        if '*' in fp_list[i]:
            wd_idx = i
            break
    assert wd_idx != -1, "The given path doesn't contain wildcard!"

    parent_folder = prefix + '/'.join(fp_list[0:wd_idx])
    if parent_folder == "" and prefix == "":
        parent_folder = "."

    wildcard = '/'.join(fp_list[wd_idx:])

    return parent_folder, wildcard

class AWSBackend:
    def __init__(self):
        self._backend = 'aws'
        self._call_prefix = ['aws', 's3']

    def copy(self, ionice, filenames, profile, quiet):
        call_args = ['ionice', '-c', '2', '-n', '7'] if ionice and (shutil.which('ionice') != None) else []
        call_args += self._call_prefix
        call_args.extend(['cp', '--only-show-errors'])
        if profile is not None:
            call_args.extend(['--profile', profile])

        source_files = filenames[:-1]
        target = filenames[-1]
        if len(source_files) > 1 and target[-1] != '/':
            target += '/'

        # Copy files one by one.
        for source in source_files:
            subcall_args = call_args.copy()
            subcall_target = target
            if '*' in source:
                # Path containing wildcard(s)
                parent_folder, wildcard = parse_wildcard(source)
                subcall_args.extend(['--recursive', '--exclude', '*', '--include', f'{wildcard}'])
                source = parent_folder + '/'
            elif source[-1] == '/': # copy an S3 folder
                subcall_args.append('--recursive')
                if subcall_target[-1] != '/':
                    subcall_target = subcall_target + '/' + os.path.basename(source)
                else:
                    subcall_target = subcall_target + os.path.basename(source)
            subcall_args.extend([source, subcall_target])
            if not quiet:
                print(' '.join(subcall_args))
            check_call(subcall_args)

    def sync(self, ionice, source, target, profile, quiet):
        call_args = ['ionice', '-c', '2', '-n', '7'] if ionice and (shutil.which('ionice') != None) else []
        call_args += self._call_prefix
        if profile is not None:
            call_args.extend(['--profile', profile])
        call_args.extend(['sync', '--delete', '--only-show-errors', source, target])
        if not quiet:
            print(' '.join(call_args))
        check_call(call_args)

    def delete(self, filenames, profile, quiet):
        call_args = self._call_prefix.copy()
        call_args.extend(['rm', '--only-show-errors'])

        if profile is not None:
            call_args.extend(['--profile', profile])

        # Delete files one by one.
        for f in filenames:
            subcall_args = call_args.copy()
            if f[-1] == '/':
                subcall_args.append('--recursive')
            subcall_args.append(f)
            if not quiet:
                print(' '.join(subcall_args))
            check_call(subcall_args)

    def stat(self, filename, profile):
        assert filename.startswith("s3://"), "Must be an S3 URI!"

        call_args = self._call_prefix.copy()
        is_folder = True if filename[-1]=='/' else False

        if is_folder:
            fn_list = filename[5:].split('/')
            bucket = fn_list[0]
            folder = '/'.join(fn_list[1:]) if len(fn_list) > 1 else ""
            session = boto3.Session(profile_name=profile)
            resp = session.client('s3').list_objects_v2(
                Bucket = bucket,
                Prefix = folder,
            )
            if resp['KeyCount'] == 0:
                raise CalledProcessError(
                    returncode=1,
                    cmd=['strato command'],
                )
        else:
            call_args.extend(['cp', '--quiet', '--dryrun', filename, '.'])
            if profile is not None:
                call_args.extend(['--profile', profile])
            check_call(call_args)



import os, shutil
from subprocess import check_call

class LocalBackend:
    def __init__(self):
        self._backend = 'local'

    def copy(self, recursive, ionice, filenames, quiet):
        assert len(filenames) >= 2, "Either source or destination is missing!"
        target = filenames[-1]
        call_args1 = ['mkdir', '-p', target]
        print(' '.join(call_args1))
        check_call(call_args1)

        call_args = ['ionice', '-c', '2', '-n', '7'] if ionice and (shutil.which('ionice')) != None else []
        call_args += ['cp']
        if recursive:
            call_args.append('-r')
        call_args.extend(filenames)
        if not quiet:
            print(' '.join(call_args))
        check_call(call_args)

    def sync(self, ionice, source, target, quiet):
        target = os.path.dirname(target)
        call_args = ['ionice', '-c', '2', '-n', '7'] if ionice and (shutil.which('ionice')) != None else []
        call_args += ['rsync', '-r', '--delete', source, target]
        if not quiet:
            print(' '.join(call_args))
        check_call(call_args)

    def delete(self, recursive, filenames, quiet):
        call_args = ['rm']
        if recursive:
            call_args.append('-r')
        call_args.extend(filenames)
        if not quiet:
            print(' '.join(call_args))
        check_call(call_args)

    def stat(self, filename):
        call_args = ['stat', filename]
        check_call(call_args)



import argparse


example_text = """Examples:
  strato exists --backend aws s3://my-bucket/file1
  strato exists --backend gcp gs://my-bucket/folder2/
  strato exists --backend local folder2/
"""

def check_status(backend, filename, profile):
    assert backend in ['aws', 'gcp', 'local'], "Backend not supported!"

    if backend == 'aws':
        from strato.backends import AWSBackend
        be = AWSBackend()
        be.stat(filename, profile=profile)
    elif backend == 'gcp':
        from strato.backends import GCPBackend
        be = GCPBackend()
        be.stat(filename)
    else:
        from strato.backends import LocalBackend
        be = LocalBackend()
        be.stat(filename)

def main(argsv):
    parser = argparse.ArgumentParser(
        description="Check the existence of a file or folder, and raise an exception if not existing. \nNotice that a folder's path must end with '/'.",
        epilog=example_text,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument('--backend', dest='backend', action='store', required=True, help='Specify which backend to use. Available options: aws, gcp, local.')
    parser.add_argument('--profile', dest='profile', type=str, action='store', help='AWS profile. Only works for aws backend, and use the default profile if not provided.')
    parser.add_argument('filename', metavar='filename', type=str, help='A file or folder path.')

    args = parser.parse_args(argsv)

    check_status(args.backend, args.filename, args.profile)



import argparse

example_text = """Examples:
  strato sync --backend aws source_folder s3://my-bucket/target_folder
  strato sync --backend gcp -m --ionice source_folder gs://my-bucket/target_folder
  strato sync --backend local source_folder target_folder
"""

def synchronize_folders(backend, parallel, ionice, source, target, profile, quiet):
    assert backend in ['aws', 'gcp', 'local'], "Backend not supported!"

    if backend == 'aws':
        from strato.backends import AWSBackend
        be = AWSBackend()
        be.sync(ionice, source, target, profile, quiet)
    elif backend == 'gcp':
        from strato.backends import GCPBackend
        be = GCPBackend()
        be.sync(parallel, ionice, source, target, quiet)
    else:
        from strato.backends import LocalBackend
        be = LocalBackend()
        be.sync(ionice, source, target, quiet)

def main(argsv):
    parser = argparse.ArgumentParser(
        description="Synchronize source and target folders.\nNotice that this synchronization deletes extra files in the target folder not found in the source folder.",
        epilog=example_text,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument('--backend', dest='backend', action='store', required=True, help='Specify which backend to use. Available options: aws, gcp, local.')
    parser.add_argument('-m', dest='parallel', action='store_true', help="Run operations in parallel. Only available for GCP backend.")
    parser.add_argument('--ionice', dest='ionice', action='store_true', help="Run with ionice to avoid monopolizing local disk's I/O. Only available for Linux.")
    parser.add_argument('--profile', dest='profile', type=str, action='store', help='AWS profile. Only works for aws backend, and use the default profile if not provided.')
    parser.add_argument('--quiet', dest='quiet', action='store_true', help="Hide the underlying cloud command.")
    parser.add_argument('source', metavar='source', type=str, help='Source folder path.')
    parser.add_argument('target', metavar='target', type=str, help='Target folder path.')

    args = parser.parse_args(argsv)
    synchronize_folders(args.backend, args.parallel, args.ionice, args.source, args.target, args.profile, args.quiet)



import argparse, sys
from strato.commands import cp, sync, rm, exists

try:
    from importlib.metadata import version
except ImportError:  # < Python 3.8: Use backport module
    from importlib_metadata import version

def main():
    command_list = [cp, sync, rm, exists]
    parser = argparse.ArgumentParser(description='Run a strato command')
    command_list_strings = list(map(lambda c: c.__name__[len('strato.commands.'):], command_list))
    parser.add_argument('command', help='The command', choices=command_list_strings)
    parser.add_argument('command_args', help='The command arguments', nargs=argparse.REMAINDER)
    parser.add_argument('-v', '--version', action="version", version=version('stratocumulus'))
 
    my_args = parser.parse_args()
    command_name = my_args.command
    command_args = my_args.command_args
    cmd = command_list[command_list_strings.index(command_name)]
    sys.argv[0] = f"strato {my_args.command}"
    cmd.main(command_args)

if __name__ == "__main__":
    main()



from setuptools import setup, find_packages
from codecs import open
from pathlib import Path
import os


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="stratocumulus",
    use_scm_version=True,
    zip_safe=False,
    description="Stratocumulus is the backend component of Cumulus.",
    long_description=long_description,
    url="https://github.com/lilab-bcb/stratocumulus",
    author="Yiming Yang, Rimte Rocher, Bo Li",
    author_email="cumulus-support@googlegroups.com",
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        'Operating System :: Microsoft :: Windows',
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Build Tools",
    ],
    keywords="Cloud computing",
    packages=find_packages(),
    setup_requires=["setuptools_scm"],
    install_requires=[
        l.strip() for l in Path("requirements.txt").read_text("utf-8").splitlines()
    ],
    python_requires=">= 3",
    entry_points={"console_scripts": ["strato=strato.__main__:main"]},
)
